import tensorflow as tf

from tensorflow_projection_qm.util import distance


@tf.function
def neighborhood_hit_impl(X_2d, y, k):
    D_low = distance.psqdist(X_2d)
    _, topk_ixs = distance.nearest_k(D_low, k)

    return tf.reduce_mean(tf.cast(tf.gather(y, topk_ixs) == y[:, tf.newaxis], tf.float64), -1)


def neighborhood_hit(X_2d, y, k):
    return tf.reduce_mean(neighborhood_hit_impl(X_2d, y, tf.constant(k)))



import numpy as np
import tensorflow as tf


@tf.function
def distance_consistency_impl(X_2d, y, n_classes):
    y_sort_ixs = tf.argsort(y)

    X_2d = tf.gather(X_2d, y_sort_ixs)  # re-order, grouped per class
    y_sorted = tf.gather(y, y_sort_ixs)
    uniq, _, sizes = tf.unique_with_counts(y_sorted)
    per_class = tf.split(X_2d, sizes, num=n_classes)

    centroids = tf.convert_to_tensor(
        [tf.reduce_mean(single_class_data, axis=0) for single_class_data in per_class]
    )

    closest_centroid = tf.argmin(
        tf.linalg.norm(tf.expand_dims(X_2d, 1) - centroids, axis=-1), axis=1
    )
    closest_centroid = tf.gather(uniq, closest_centroid)

    return tf.reduce_mean(tf.cast(closest_centroid == y_sorted, tf.float64))


def distance_consistency(X_2d, y):
    n_classes = len(np.unique(y))

    return distance_consistency_impl(X_2d, y, n_classes)



import tensorflow as tf

from tensorflow_projection_qm.util import distance


@tf.function
def continuity_impl(X, X_2d, k) -> tf.Tensor:
    k = tf.cast(k, tf.int32)
    D_high = distance.psqdist(X)
    D_low = distance.psqdist(X_2d)

    n = tf.shape(D_high)[0]

    nn_orig = distance.sort_distances(D_high)
    nn_proj = distance.sort_distances(D_low)
    ixs_proj = tf.argsort(nn_proj)

    knn_orig = nn_orig[:, 1 : k + 1]
    knn_proj = nn_proj[:, 1 : k + 1]

    V_i = tf.sparse.to_dense(tf.sets.difference(knn_orig, knn_proj), default_value=-1)
    pre_cont = tf.where(V_i != -1, tf.gather(ixs_proj, V_i, batch_dims=-1) - k, 0)
    cont = tf.reduce_sum(pre_cont, -1)
    cont_t = tf.cast(cont, tf.float64)
    k = tf.cast(k, tf.float64)
    n = tf.cast(n, tf.float64)
    return tf.squeeze(1 - (2 / (k * (2 * n - 3 * k - 1)) * cont_t))


def continuity(X, X_2d, k: int) -> tf.Tensor:
    return tf.reduce_mean(continuity_impl(X, X_2d, tf.constant(k)))



from .continuity import continuity
from .distance_consistency import distance_consistency
from .mean_rel_rank_error import mrre_data, mrre_proj
from .neighborhood_hit import neighborhood_hit
from .stress import (
    normalized_stress,
    normalized_stress_from_distances,
    raw_stress,
    raw_stress_from_distances,
    scale_normalized_stress,
    scale_normalized_stress_from_distances,
    scaled_stress,
    scaled_stress_from_distances,
)
from .trustworthiness import trustworthiness

__all__ = [
    "continuity",
    "distance_consistency",
    "neighborhood_hit",
    "mrre_data",
    "mrre_proj",
    "normalized_stress",
    "normalized_stress_from_distances",
    "raw_stress",
    "raw_stress_from_distances",
    "scale_normalized_stress",
    "scale_normalized_stress_from_distances",
    "scaled_stress",
    "scaled_stress_from_distances",
    "trustworthiness",
]
